#include <assert.h>
#include "io.h"
#include "objects.h"
#include "map.h"
#include "misc.h"
#include "monster.h"
#include "player.h"

map *maps[64];
map *m;

typedef struct
{ int xstart, ystart, xsize, ysize;
} room;

coord diradd[4]=
{ { -1,  0 },
  {  0, -1 },
  {  1,  0 },
  {  0,  1 }
};

static int makerooms(int rooms, int space);
static void makecorridors(int nrooms);
static void makedoors();
static void makestairs(int up, int down);

map *makemap(int xsize, int ysize, int nrooms, int space, int up, int down)
{ int i;

  m->place=zcalloc(xsize, sizeof(block *));
  for(i=0; i<xsize; i++) m->place[i]=zcalloc(ysize, sizeof(block));
  m->lastseen=zcalloc(xsize, sizeof(char *));
  for(i=0; i<xsize; i++) m->lastseen[i]=zmalloc(ysize);

  m->mhead=(monster *)zmalloc(sizeof(monster));
  m->xsize=xsize;
  m->ysize=ysize;

  makecorridors(makerooms(nrooms, space));
  makedoors();
  makestairs(up, down);

  return m;
}

coord findfreespace()
{ coord c;

  do
  { c.x=randint(m->xsize);
    c.y=randint(m->ysize);
  } while(!can_go(c) ||
          here(c.x, c.y)==STAIRWAY_UP || here(c.x, c.y)==STAIRWAY_DOWN);

  return c;
}

coord findfixture(int fixture)
{ int i, j;

  for(i=0; i<m->xsize; i++)
    for(j=0; j<m->ysize; j++)
      if(here(i, j)==fixture)
        return (coord){ i, j };

  return (coord){ 0, 0 };
}

int walkable(coord c)
{ if(here(c.x, c.y)==WALL) return 0;
  if(here(c.x, c.y)==DOOR && special(c.x, c.y).closed) return 0;
  return 1;
}

static room randroom()
{ room r;

  do
  { r.xstart=randint(m->xsize-9)+3;
    r.ystart=randint(m->ysize-9)+3;
    r.xsize=randint(6)+4;
    r.ysize=randint(6)+4;
  } while(r.xstart+r.xsize>=m->xsize-2 || r.ystart+r.ysize>=m->ysize-2);

  return r;
}

static void addroom(room r, int nroom)
{ int i, j;

  for(i=0; i<r.xsize; i++)
    for(j=0; j<r.ysize; j++)
      here(r.xstart+i, r.ystart+j)=FLOOR+nroom;
}

static int overlap(room r, room r2, int allow)
{ return r.xstart+r.xsize>r2.xstart+allow &&
         r.xstart<r2.xstart+r2.xsize-allow &&
         r.ystart+r.ysize>r2.ystart+allow &&
         r.ystart<r2.ystart+r2.ysize-allow;
}

static int makerooms(int rooms, int space)
{ int fail=0, nrooms=0, i;
  room roomlist[rooms];
  room r;

  do
  { int failure=0;
    do
    { r=randroom();
      for(i=fail=0; i<nrooms; i++) fail+=overlap(r, roomlist[i], space);
      if(failure++>50) return nrooms;
    } while(fail);
    addroom(r, nrooms);
    roomlist[nrooms++]=r;
  } while(nrooms<rooms);

  return nrooms;
}

static int adjacent(coord c, int type)
{ return (here(c.x-1, c.y)==type)+(here(c.x+1, c.y)==type)+
         (here(c.x, c.y-1)==type)+(here(c.x, c.y+1)==type);
}

static coord findroom(int type)
{ coord c;

  do
  { c.x=randint(m->xsize-2)+1;
    c.y=randint(m->ysize-2)+1;
  } while(!adjacent(c, type) || here(c.x, c.y)!=WALL);

  return c;
}

static int adjnot(coord c, int room)
{ int i;

  for(i=0; i<4; i++)
  { char w=here(c.x+diradd[i].x, c.y+diradd[i].y);
    if(w!=WALL && w!=room) return w;
  }

  return 0;
}

static void renumber(int old, int new)
{ int i, j;

  for(i=0; i<m->xsize; i++)
    for(j=0; j<m->ysize; j++)
      if(here(i, j)==old) here(i, j)=new;
}

static int diagadjacent(coord c)
{ return (here(c.x-1, c.y-1) && !(here(c.x-1, c.y) || here(c.x, c.y-1))) ||
         (here(c.x+1, c.y-1) && !(here(c.x+1, c.y) || here(c.x, c.y-1))) ||
         (here(c.x-1, c.y+1) && !(here(c.x-1, c.y) || here(c.x, c.y+1))) ||
         (here(c.x+1, c.y+1) && !(here(c.x+1, c.y) || here(c.x, c.y+1)));
}

static int dig(coord c, int room)
{ coord new;
  int fails=0;
  int dir, dir2;

  here(c.x, c.y)=127;

  dir=randint(4);
  dir2=randint(2)?dir+1:dir-1;
  if(dir2<0) dir2=3;
  if(dir2>3) dir2=0;

  for(;;)
  { new=c;
    if(!randint(5)) swap(dir, dir2);
    new.x+=diradd[dir].x;
    new.y+=diradd[dir].y;
    if(new.x<2 || new.y<2 || new.x>m->xsize-2 || new.y>m->ysize-2 ||
       adjacent(new, 127)+adjacent(new, room)>1 ||
       diagadjacent(new))
    { fails++;
      swap(dir, dir2);
    }
    else
    { fails=0;
      c=new;
      here(c.x, c.y)=127;
      if(adjacent(c, WALL)<3)
      { renumber(127, room);
        return adjnot(c, room);
      }
    }
    if(fails==2) return 0;
  }
}

static void makecorridors(int nrooms)
{ int n;

  while(nrooms>1)
  { do
    { if(!(n=dig(findroom(nrooms), nrooms)))
        renumber(127, WALL);
    } while(!n);
    renumber(nrooms, n);
    nrooms--;
  }
}

#define DOORWAY(x, y, rotate) \
 (here((x), (y)-1)==((rotate)?FLOOR:WALL) && \
  here((x), (y)+1)==((rotate)?FLOOR:WALL) && \
  here((x)-1, (y))==((rotate)?WALL:FLOOR) && \
  here((x)+1, (y))==((rotate)?WALL:FLOOR) && \
  ((here((x)+1, (y)+1)==FLOOR && here((x)+1, (y)-1)==FLOOR) || \
   (here((x)+1, (y)-1)==FLOOR && here((x)-1, (y)-1)==FLOOR) || \
   (here((x)-1, (y)+1)==FLOOR && here((x)-1, (y)-1)==FLOOR) || \
   (here((x)-1, (y)+1)==FLOOR && here((x)+1, (y)+1)==FLOOR)))

static void makedoors()
{ int i, j;

  for(i=1; i<m->xsize-1; i++)
    for(j=1; j<m->ysize-1; j++)
      if(here(i, j)==FLOOR && (DOORWAY(i, j, 0) || DOORWAY(i, j, 1)) &&
         !randint(2))
      { here(i, j)=DOOR;
        special(i, j).closed=randint(3);
      }
}

static void makestairs(int up, int down)
{ coord c;

  c=findfreespace();
  here(c.x, c.y)=STAIRWAY_DOWN;
  special(c.x, c.y).destination=down;
  if(up!=-1)
  { c=findfreespace();
    here(c.x, c.y)=STAIRWAY_UP;
    special(c.x, c.y).destination=up;
  }
}

void fillmap(int nobjs, int nmonsts)
{ int i;

  for(i=0; i<nobjs; i++)
    addobjtotile(findfreespace(), createobject(randobject()));

  for(i=0; i<nmonsts; i++)
    addmonster(createmonster(randmonster(pstuff.level), findfreespace()));
}

void changelevel(int n, int up, int down, int where)
{ pstuff.level=n;

  m=maps[pstuff.level];
  if(!m)
  { m=maps[pstuff.level]=zmalloc(sizeof(map));
    makemap(SCREEN_WIDTH, SCREEN_HEIGHT-2, 20, -3, up, down);
    fillmap(roll((die) { 5, 3, 0 }), roll((die) { 5, 3, 0 }));
  }

  assert(m);
  you.where=findfixture(where);
}
